@@ -37,36 +37,55 @@ def map_valid(data_field, map_field, result=None):
 
 
 def ordered_map_valid_stream(data_field, map_field, result_field, chunksize=DEFAULT_CHUNKSIZE):
-    df_it = iter(chunks(len(data_field.data)))
-    mf_it = iter(chunks(len(map_field.data)))
+    df_it = iter(chunks(len(data_field.data), chunksize=chunksize))
+    mf_it = iter(chunks(len(map_field.data), chunksize=chunksize))
     df_range = next(df_it)
     mf_range = next(mf_it)
     dfc = data_field.data[df_range[0]:df_range[1]]
     mfc = map_field.data[mf_range[0]:mf_range[1]]
-    rslt = np.zeros(chunksize, dtype=data_field.data.dtype)
 
-    i = 0
-    while i < len(map_field.data):
-        ii, d = ordered_map_valid_partial(dfc, mfc, rslt)
+    is_field_parameter = val.is_field_parameter(result_field)
+    result_dtype = result_field.data.dtype if is_field_parameter else result_field.dtype
+    rslt = np.zeros(chunksize, dtype=result_dtype)
+
+    m = 0
+    d = 0
+    while m < len(map_field.data):
+        mm, dd = ordered_map_valid_partial(df_range[0], dfc, mfc, rslt)
+        if mm > 0:
+            if is_field_parameter:
+                result_field.data.write(rslt[:mm])
+            else:
+                result_field[m:m + mm] = rslt[:mm]
+            rslt[:] = 0
+        m += mm
+        if m == mf_range[1] and m < len(map_field.data):
+            mf_range = next(mf_it)
+            mfc = map_field.data[mf_range[0]:mf_range[1]]
+        else:
+            mfc = mfc[mm:]
+        if dd >= df_range[1] and dd < len(data_field.data):
+            df_range = next(df_it)
+            dfc = data_field.data[df_range[0]:df_range[1]]
+        # else:
+        #     dfc = dfc[dd:]
 
 
 
 # 0 2 3 4 5 7 8 9 11 12 14 15 17 18 19
 # 0 0 0 0 1 1 1 1  2  2  2  2  3  3  3
-def ordered_map_valid_partial(data_field, map_field, result):
+def ordered_map_valid_partial(d, data_field, map_field, result):
     i = 0
-    while i < len(map_field):
+    while True:
         val = map_field[i]
         if val < INVALID_INDEX:
-            if val >= len(data_field):
+            if val >= d + len(data_field):
                 # need a new data_field chunk
                 return i, val
-            result[i] = val
+            result[i] = data_field[val - d]
         i += 1
-    return i, len(data_field)
-
-
-
+        if i >= len(map_field):
+            return i, val
 
 
 @njit
@@ -371,7 +390,74 @@ def apply_spans_concat(spans, src_index, src_values, dest_index, dest_values,
     return s+1, index_i, index_v
 
 
-def ordered_map_to_right_left_unique_streamed(left, right, left_to_right, chunksize=1 << 20):
+# def ordered_map_to_right_left_unique_streamed(left, right, left_to_right, chunksize=1 << 20):
+#     i = 0
+#     j = 0
+#     lc_it = iter(chunks(len(left.data), chunksize))
+#     lc_range = next(lc_it)
+#     rc_it = iter(chunks(len(right.data), chunksize))
+#     rc_range = next(rc_it)
+#     lc = left.data[lc_range[0]:lc_range[1]]
+#     rc = right.data[rc_range[0]:rc_range[1]]
+#     acc_written = 0
+#
+#     ltri = np.zeros(chunksize, dtype=left_to_right.data.dtype)
+#     unmapped = 0
+#     while i < len(left.data) and j < len(right.data):
+#         ii, jj, u = ordered_map_to_right_left_unique_partial(i, lc, rc, ltri)
+#         unmapped += u
+#         if jj > 0:
+#             if val.is_field_parameter(left_to_right):
+#                 left_to_right.data.write(ltri[:jj])
+#             else:
+#                 left_to_right[acc_written:acc_written + jj] = ltri[:jj]
+#                 acc_written += jj
+#         i += ii
+#         j += jj
+#         if i > lc_range[1]:
+#             raise ValueError("'i' has got ahead of current chunk; this shouldn't happen")
+#         if j > rc_range[1]:
+#             raise ValueError("'j' has got ahead of current chunk; this shouldn't happen")
+#         if i == lc_range[1] and i < len(left.data):
+#             lc_range = next(lc_it)
+#             lc = left.data[lc_range[0]:lc_range[1]]
+#         else:
+#             lc = lc[ii:]
+#         if j == rc_range[1] and j < len(right.data):
+#             rc_range = next(rc_it)
+#             rc = right.data[rc_range[0]:rc_range[1]]
+#         else:
+#             rc = rc[jj:]
+#     return unmapped > 0
+
+
+# @njit
+# def ordered_map_to_right_left_unique_partial(d_i, left, right, left_to_right):
+#     """
+#     Returns:
+#     [0]: how many positions forward i moved
+#     [1]: how many positions forward j moved
+#     [2]: how many elements were written
+#     """
+#     i = 0
+#     j = 0
+#     unmapped = 0
+#     while i < len(left) and j < len(right):
+#         if left[i] < right[j]:
+#             i += 1
+#         elif left[i] > right[j]:
+#             left_to_right[j] = INVALID_INDEX
+#             j += 1
+#             unmapped += 1
+#         else:
+#             left_to_right[j] = i + d_i
+#             if j+1 < len(right) and right[j+1] != right[j]:
+#                 i += 1
+#             j += 1
+#     return i, j, unmapped
+
+
+def ordered_map_to_right_right_unique_streamed(left, right, left_to_right, chunksize=1 << 20):
     i = 0
     j = 0
     lc_it = iter(chunks(len(left.data), chunksize))
@@ -382,17 +468,19 @@ def ordered_map_to_right_left_unique_streamed(left, right, left_to_right, chunks
     rc = right.data[rc_range[0]:rc_range[1]]
     acc_written = 0
 
-    ltri = np.zeros(chunksize, dtype=left_to_right.data.dtype)
+    is_field_parameter = val.is_field_parameter(left_to_right)
+    result_dtype = left_to_right.data.dtype if is_field_parameter else left_to_right.dtype
+    ltri = np.zeros(chunksize, dtype=result_dtype)
     unmapped = 0
     while i < len(left.data) and j < len(right.data):
-        ii, jj, u = ordered_map_to_right_left_unique_partial(i, lc, rc, ltri)
+        ii, jj, u = ordered_map_to_right_right_unique_partial(j, lc, rc, ltri)
         unmapped += u
-        if jj > 0:
-            if val.is_field_parameter(left_to_right):
-                left_to_right.data.write(ltri[:jj])
+        if ii > 0:
+            if is_field_parameter:
+                left_to_right.data.write(ltri[:ii])
             else:
-                left_to_right[acc_written:acc_written + jj] = ltri[:jj]
-                acc_written += jj
+                left_to_right[acc_written:acc_written + ii] = ltri[:ii]
+                acc_written += ii
         i += ii
         j += jj
         if i > lc_range[1]:
@@ -411,52 +499,56 @@ def ordered_map_to_right_left_unique_streamed(left, right, left_to_right, chunks
             rc = rc[jj:]
     return unmapped > 0
 
-"""
-Returns:
-[0]: how many positions forward i moved
-[1]: how many positions forward j moved
-[2]: how many elements were written
-"""
+
 @njit
-def ordered_map_to_right_left_unique_partial(d_i, left, right, left_to_right):
+def ordered_map_to_right_right_unique_partial(d_j, left, right, left_to_right):
+    """
+    Returns:
+    [0]: how many positions forward i moved
+    [1]: how many positions forward j moved
+    [2]: how many elements were written
+    """
     i = 0
     j = 0
     unmapped = 0
     while i < len(left) and j < len(right):
         if left[i] < right[j]:
+            left_to_right[i] = INVALID_INDEX
             i += 1
-        elif left[i] > right[j]:
-            left_to_right[j] = INVALID_INDEX
-            j += 1
             unmapped += 1
-        else:
-            left_to_right[j] = i + d_i
-            if j+1 < len(right) and right[j+1] != right[j]:
-                i += 1
+        elif left[i] > right[j]:
             j += 1
+        else:
+            left_to_right[i] = j + d_j
+            if i+1 >= len(left) or left[i+1] != left[i]:
+                j += 1
+            i += 1
+            # if j+1 < len(right) and right[j+1] != right[j]:
+            #     i += 1
+            # j += 1
     return i, j, unmapped
 
 
 @njit
-def ordered_map_to_right_left_unique(first, second, result):
-    if len(second) != len(result):
-        msg = "'second' and 'result' must be the same length"
+def ordered_map_to_right_right_unique(first, second, result):
+    if len(first) != len(result):
+        msg = "'first' and 'result' must be the same length"
         raise ValueError(msg)
     i = 0
     j = 0
     unmapped = 0
     while i < len(first) and j < len(second):
         if first[i] < second[j]:
+            result[i] = INVALID_INDEX
             i += 1
-        elif first[i] > second[j]:
-            result[j] = INVALID_INDEX
-            j += 1
             unmapped += 1
-        else:
-            result[j] = i
-            if j+1 < len(second) and second[j+1] != second[j]:
-                i += 1
+        elif first[i] > second[j]:
             j += 1
+        else:
+            result[i] = j
+            if i+1 >= len(first) or first[i+1] != first[i]:
+                j += 1
+            i += 1
 
     while j < len(second):
         result[j] = INVALID_INDEX
@@ -467,7 +559,7 @@ def ordered_map_to_right_left_unique(first, second, result):
 
 @njit
 def ordered_map_to_right_both_unique(first, second, result):
-    if len(second) != len(result):
+    if len(first) != len(result):
         msg = "'second' and 'result' must be the same length"
         raise ValueError(msg)
     i = 0
@@ -476,19 +568,19 @@ def ordered_map_to_right_both_unique(first, second, result):
     unmapped = 0
     while i < len(first) and j < len(second):
         if first[i] < second[j]:
+            result[i] = INVALID_INDEX
             i += 1
-        elif first[i] > second[j]:
-            result[j] = INVALID_INDEX
-            j += 1
             unmapped += 1
+        elif first[i] > second[j]:
+            j += 1
         else:
-            result[j] = i
+            result[i] = j
             i += 1
             j += 1
 
-    while j < len(second):
-        result[j] = INVALID_INDEX
-        j += 1
+    while i < len(first):
+        result[i] = INVALID_INDEX
+        i += 1
 
     return unmapped > 0
 
@@ -575,15 +667,15 @@ def ordered_inner_map_left_unique_streamed(left, right, left_to_inner, right_to_
         else:
             rc = rc[jj:]
 
-"""
-Returns:
-[0]: how many positions forward i moved
-[1]: how many positions forward j moved
-[2]: how many elements were written
-"""
 @njit
 def ordered_inner_map_left_unique_partial(d_i, d_j, left, right,
                                           left_to_inner, right_to_inner):
+    """
+    Returns:
+    [0]: how many positions forward i moved
+    [1]: how many positions forward j moved
+    [2]: how many elements were written
+    """
     i = 0
     j = 0
     m = 0
@@ -597,7 +689,7 @@ def ordered_inner_map_left_unique_partial(d_i, d_j, left, right,
             left_to_inner[m] = i + d_i
             right_to_inner[m] = j + d_j
             m += 1
-            if j+1 < len(right) and right[j+1] != right[j]:
+            if j+1 >= len(right) or right[j+1] != right[j]:
                 i += 1
             j += 1
     return i, j, m
