@@ -193,7 +193,7 @@ class Session:
         """
         filter_to_apply_ = val.array_from_parameter(self, 'index_to_apply', filter_to_apply)
         writer_ = None
-        if writer_:
+        if dest is not None:
             writer_ = val.field_from_parameter(self, 'writer', dest)
         if isinstance(src, fld.IndexedStringField):
             src_ = val.field_from_parameter(self, 'reader', src)
@@ -201,13 +201,14 @@ class Session:
                 ops.apply_filter_to_index_values(filter_to_apply_,
                                                  src_.indices[:], src_.values[:])
             if writer_:
-                writer_.write_raw(dest_indices, dest_values)
+                writer_.indices.write(dest_indices)
+                writer_.values.write(dest_values)
             return dest_indices, dest_values
         else:
             reader_ = val.array_from_parameter(self, 'reader', src)
             result = reader_[filter_to_apply]
             if writer_:
-                writer_.write(result)
+                writer_.data.write(result)
             return result
 
 
@@ -224,7 +225,7 @@ class Session:
         """
         index_to_apply_ = val.array_from_parameter(self, 'index_to_apply', index_to_apply)
         writer_ = None
-        if writer_:
+        if dest is not None:
             writer_ = val.field_from_parameter(self, 'writer', dest)
         if isinstance(src, fld.IndexedStringField):
             src_ = val.field_from_parameter(self, 'reader', src)
@@ -232,13 +233,14 @@ class Session:
                 ops.apply_indices_to_index_values(index_to_apply_,
                                                   src_.indices[:], src_.values[:])
             if writer_:
-                writer_.write_raw(dest_indices, dest_values)
+                writer_.indices.write(dest_indices)
+                writer_.values.write(dest_values)
             return dest_indices, dest_values
         else:
             reader_ = val.array_from_parameter(self, 'reader', src)
             result = reader_[index_to_apply]
             if writer_:
-                writer_.write(result)
+                writer_.data.write(result)
             return result
 
 
@@ -294,7 +296,6 @@ class Session:
                 raw_fields.append(f[:] if isinstance(f, rw.Reader) else f)
         return per._get_spans(raw_field, raw_fields)
 
-
     def _apply_spans_no_src(self, predicate, spans, dest=None):
         if dest:
             if val.is_field_parameter(self, 'dest', dest):
@@ -316,10 +317,9 @@ class Session:
             predicate(spans, results)
             return results
 
-
     def _apply_spans_src(self, predicate, spans, src, dest=None):
-        src_ = val.array_from_parameter(src)
-        if len(src) != len(spans) - 1:
+        src_ = val.array_from_parameter(self, 'src', src)
+        if len(src) != spans[-1]:
             error_msg = ("'src' (length {}) must be one element shorter than 'spans' "
                          "(length {})")
             raise ValueError(error_msg.format(len(src_), len(spans)))
@@ -337,11 +337,11 @@ class Session:
                     error_msg = ("if 'dest' (length {}) is an ndarray, it must be one element "
                                  "shorter than the length of 'spans' (length{})")
                     raise ValueError(error_msg.format(len(dest_), len(spans)))
-                    predicate(spans, dest_)
+                    predicate(spans, src, dest_)
                 return dest_
         else:
-            results = np.zeros(len(spans) - 1, dtype='int64')
-            predicate(spans, results)
+            results = np.zeros(len(spans) - 1, dtype=src_.dtype)
+            predicate(spans, src, results)
             return results
 
     def apply_spans_index_of_min(self, spans, src, dest=None):
@@ -365,11 +365,11 @@ class Session:
     def apply_spans_max(self, spans, src, dest=None):
         return self._apply_spans_src(ops.apply_spans_max, spans, src, dest)
 
-    def apply_spans_first(self, spans, dest=None):
-        return self._apply_spans_no_src(ops.apply_spans_first, spans, dest)
+    def apply_spans_first(self, spans, src, dest=None):
+        return self._apply_spans_src(ops.apply_spans_first, spans, src, dest)
 
-    def apply_spans_last(self, spans, dest=None):
-        return self._apply_spans_no_src(ops.apply_spans_last, spans, dest)
+    def apply_spans_last(self, spans, src, dest=None):
+        return self._apply_spans_src(ops.apply_spans_last, spans, src, dest)
 
     def apply_spans_concat(self, spans, src, dest):
         if not isinstance(src, fld.IndexedStringField):
@@ -561,6 +561,9 @@ class Session:
         writer.write(destination_space_values)
 
     def get(self, field):
+        if isinstance(field, fld.Field):
+            return field
+
         if 'fieldtype' not in field.attrs.keys():
             raise ValueError(f"'{field}' is not a well-formed field")
 
@@ -777,8 +780,7 @@ class Session:
 
 
     def merge_left(self, left_on, right_on,
-                   left_fields=tuple(), right_fields=tuple(),
-                   left_writers=None, right_writers=None):
+                   right_fields=tuple(), right_writers=None):
         l_key_raw = val.raw_array_from_parameter(self, 'left_on', left_on)
         l_index = np.arange(len(l_key_raw), dtype=np.int64)
         l_df = pd.DataFrame({'l_k': l_key_raw, 'l_index': l_index})
@@ -788,43 +790,23 @@ class Session:
         r_df = pd.DataFrame({'r_k': r_key_raw, 'r_index': r_index})
 
         df = pd.merge(left=l_df, right=r_df, left_on='l_k', right_on='r_k', how='left')
-        l_to_r_map = df['l_index'].to_numpy()
-        l_to_r_filt = np.logical_not(df['l_index'].isnull()).to_numpy()
         r_to_l_map = df['r_index'].to_numpy(dtype=np.int64)
         r_to_l_filt = np.logical_not(df['r_index'].isnull()).to_numpy()
-        # print(df)
-
-        # print("l_to_r_map:", l_to_r_map, l_to_r_map.dtype)
-        # print("r_to_l_map:", r_to_l_map, r_to_l_map.dtype)
-
-        # print(r_to_l_map > -1)
-        left_results = list()
-        for ilf, lf in enumerate(left_fields):
-            lf_raw = val.raw_array_from_parameter(self, 'left_fields[{}]'.format(ilf), lf)
-            joined_field = per._safe_map(lf_raw, l_to_r_map, l_to_r_filt)
-            print(joined_field)
-            if left_writers == None:
-                left_results.append(joined_field)
-            else:
-                left_writers[ilf].write(joined_field)
 
         right_results = list()
         for irf, rf in enumerate(right_fields):
             rf_raw = val.raw_array_from_parameter(self, 'right_fields[{}]'.format(irf), rf)
             joined_field = per._safe_map(rf_raw, r_to_l_map, r_to_l_filt)
-            print(joined_field)
-            if right_writers == None:
+            if right_writers is None:
                 right_results.append(joined_field)
             else:
-                right_writers[irf].write(joined_field)
+                right_writers[irf].data.write(joined_field)
 
-        if left_writers == None:
-            return left_results, right_results
+        return right_results
 
 
     def merge_right(self, left_on, right_on,
-                    left_fields=None, right_fields=None,
-                    left_writers=None, right_writers=None):
+                    left_fields=None, left_writers=None):
         l_key_raw = val.raw_array_from_parameter(self, 'left_on', left_on)
         l_index = np.arange(len(l_key_raw), dtype=np.int64)
         l_df = pd.DataFrame({'l_k': l_key_raw, 'l_index': l_index})
@@ -834,33 +816,19 @@ class Session:
         r_df = pd.DataFrame({'r_k': r_key_raw, 'r_index': r_index})
 
         df = pd.merge(left=r_df, right=l_df, left_on='r_k', right_on='l_k', how='left')
-        l_to_r_map = df['r_index'].to_numpy()
-        l_to_r_filt = np.logical_not(df['r_index'].isnull())
-        r_to_l_map = df['l_index'].to_numpy(dtype=np.int64)
-        r_to_l_filt = np.logical_not(df['l_index'].isnull())
-        print(df)
+        l_to_r_map = df['l_index'].to_numpy(dtype='int64')
+        l_to_r_filt = np.logical_not(df['l_index'].isnull()).to_numpy()
 
-        print("l_to_r_map:", l_to_r_map, l_to_r_map.dtype)
-        print("r_to_l_map:", r_to_l_map, r_to_l_map.dtype)
-
-        print(r_to_l_map > -1)
         left_results = list()
         for ilf, lf in enumerate(left_fields):
-            joined_field = per._safe_map(lf, r_to_l_map)
-            print(joined_field)
-            if left_writers == None:
+            lf_raw = val.raw_array_from_parameter(self, 'left_fields[{}]'.format(ilf), lf)
+            joined_field = per._safe_map(lf_raw, l_to_r_map, l_to_r_filt)
+            if left_writers is None:
                 left_results.append(joined_field)
             else:
                 left_writers[ilf].write(joined_field)
 
-        right_results = list()
-        for irf, rf in enumerate(right_fields):
-            joined_field = per._safe_map(rf, l_to_r_map)
-            print(joined_field)
-            if right_writers == None:
-                right_results.append(joined_field)
-            else:
-                right_writers[irf].write(joined_field)
+        return left_results
 
 
     def merge_inner(self, left_on, left_fields, right_on, right_fields):
@@ -901,7 +869,7 @@ class Session:
                 src_ = val.array_from_parameter(self, 'left_field_sources', src)
                 snk_ = val.array_from_parameter(self, 'left_field_sinks', snk)
                 ops.map_valid(src_, field_map, snk_)
-        return rtn_sinks
+        return None if rtn_sinks is None else tuple(rtn_sinks)
 
 
     def ordered_merge(self, left_on, right_on, how,
@@ -928,7 +896,28 @@ class Session:
     def ordered_left_merge(self, left_on, right_on, left_to_right_map,
                            left_field_sources=tuple(), left_field_sinks=None,
                            left_unique=False, right_unique=False):
-
+        """
+        Generate the results of a left join apply it to the fields described in the tuple
+        'left_field_sources'. If 'left_field_sinks' is set, the mapped values are written
+        to the fields / arrays set there.
+        Note: in order to achieve best scalability, you should use groups / fields rather
+        than numpy arrays and provide a tuple of groups/fields to left_field_sinks, so
+        that the session and compute the merge and apply the mapping in a streaming
+        fashion.
+        :param left_on: the group/field/numba array that contains the left key values
+        :param right_on: the group/field/numba array that contains the right key values
+        :param left_to_right_map: a group/field/numba array that the map is written to. If
+        it is a numba array, it must be the size of the resulting merge
+        :param left_field_sources: a tuple of group/fields/numba arrays that contain the
+        fields to be joined
+        :param left_field_sinks: optional - a tuple of group/fields/numba arrays that
+        the mapped fields should be written to
+        :param left_unique: a hint to indicate whether the 'left_on' field contains unique
+        values
+        :param right_unique: a hint to indicate whether the 'right_on' field contains
+        unique values
+        :return: If left_field_sinks is not set, a tuple of the output fields is returned
+        """
         if left_field_sinks is not None:
             if len(left_field_sources) != len(left_field_sinks):
                 msg = ("{} and {} should be of the same length but are length {} and {} "
@@ -949,23 +938,31 @@ class Session:
             has_unmapped = None
             if left_unique == False:
                 if right_unique == False:
-                    raise NotImplementedError()
+                    raise ValueError("Right key must not have duplicates")
                 else:
-                    raise NotImplementedError()
-            else:
-                if right_unique == False:
                     if streamable:
-                        has_unmapped =\
-                            ops.ordered_map_to_right_left_unique_streamed(left_on, right_on,
+                        has_unmapped = \
+                            ops.ordered_map_to_right_right_unique_streamed(left_on, right_on,
                                                                           left_to_right_map)
                         result = left_to_right_map.data[:]
                     else:
-                        result = np.zeros(len(right_on), dtype=np.int64)
-                        has_unmapped =\
-                            ops.ordered_map_to_right_left_unique(left_on, right_on, result)
-
+                        result = np.zeros(len(left_on), dtype=np.int64)
+                        has_unmapped = \
+                            ops.ordered_map_to_right_right_unique(left_on, right_on, result)
+            else:
+                if right_unique == False:
+                    raise ValueError("Right key must not have duplicates")
+                    # if streamable:
+                    #     has_unmapped =\
+                    #         ops.ordered_map_to_right_left_unique_streamed(left_on, right_on,
+                    #                                                       left_to_right_map)
+                    #     result = left_to_right_map.data[:]
+                    # else:
+                    #     result = np.zeros(len(left_on), dtype=np.int64)
+                    #     has_unmapped =\
+                    #         ops.ordered_map_to_right_left_unique(left_on, right_on, result)
                 else:
-                    result = np.zeros(len(right_on), dtype=np.int64)
+                    result = np.zeros(len(left_on), dtype=np.int64)
                     has_unmapped = ops.ordered_map_to_right_both_unique(left_on, right_on, result)
 
         with utils.Timer("mapping fields", new_line=True):
@@ -973,10 +970,32 @@ class Session:
         return rtn_left_sinks
 
 
-    def ordered_right_merge(self, left_on, right_on,
+    def ordered_right_merge(self, left_on, right_on, right_to_left_map,
                             right_field_sources=tuple(), right_field_sinks=None,
                             left_unique=False, right_unique=False):
-        return self.ordered_left_merge(right_on, left_on,
+        """
+        Generate the results of a right join apply it to the fields described in the tuple
+        'right_field_sources'. If 'right_field_sinks' is set, the mapped values are written
+        to the fields / arrays set there.
+        Note: in order to achieve best scalability, you should use groups / fields rather
+        than numpy arrays and provide a tuple of groups/fields to right_field_sinks, so
+        that the session and compute the merge and apply the mapping in a streaming
+        fashion.
+        :param left_on: the group/field/numba array that contains the left key values
+        :param right_on: the group/field/numba array that contains the right key values
+        :param right_to_left_map: a group/field/numba array that the map is written to. If
+        it is a numba array, it must be the size of the resulting merge
+        :param right_field_sources: a tuple of group/fields/numba arrays that contain the
+        fields to be joined
+        :param right_field_sinks: optional - a tuple of group/fields/numba arrays that
+        the mapped fields should be written to
+        :param left_unique: a hint to indicate whether the 'left_on' field contains unique
+        values
+        :param right_unique: a hint to indicate whether the 'right_on' field contains
+        unique values
+        :return: If right_field_sinks is not set, a tuple of the output fields is returned
+        """
+        return self.ordered_left_merge(right_on, left_on, right_to_left_map,
                                        right_field_sources, right_field_sinks,
                                        right_unique, left_unique)
 
